@@ -11,6 +11,8 @@
                             Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc>,
                             Leonhard Wimmer <leo@mediatomb.cc>
 
+    Copyright (C) 2016-2023 Gerbera Contributors
+
     MediaTomb is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License version 2
     as published by the Free Software Foundation.
@@ -28,12 +30,12 @@
 */
 
 /// \file server.cc
+#define LOG_FAC log_facility_t::server
 
 #include "server.h" // API
 
 #ifdef HAVE_CURL
 #include <curl/curl.h>
-#include <iostream>
 #endif
 
 #include <upnpconfig.h>
@@ -41,32 +43,57 @@
 #include <upnptools.h>
 #endif
 
-#include <chrono>
+#include <sstream>
 #include <thread>
 
+#include "action_request.h"
 #include "config/config_manager.h"
 #include "content/content_manager.h"
 #include "database/database.h"
 #include "device_description_handler.h"
 #include "file_request_handler.h"
-#include "serve_request_handler.h"
+#include "iohandler/io_handler.h"
+#include "subscription_request.h"
+#include "upnp_cds.h"
+#include "upnp_cm.h"
+#include "upnp_common.h"
+#include "upnp_mrreg.h"
+#include "util/grb_net.h"
 #include "util/mime.h"
+#include "util/tools.h"
 #include "util/upnp_clients.h"
+#include "util/url_utils.h"
+#include "util/xml_to_json.h"
 #include "web/pages.h"
 #include "web/session_manager.h"
 
+#ifdef HAVE_JS
+#include "content/scripting/playlist_parser_script.h"
+#endif
+
 #ifdef HAVE_CURL
 #include "url_request_handler.h"
 #endif
 
+#ifdef HAVE_INOTIFY
+#include "content/autoscan_inotify.h"
+#endif
+
+#ifdef ONLINE_SERVICES
+#include "content/onlineservice/online_service.h"
+#endif
+
+constexpr auto DEVICE_DESCRIPTION_PATH = std::string_view("description.xml");
+
 Server::Server(std::shared_ptr<Config> config)
     : config(std::move(config))
 {
 }
 
-void Server::init()
+void Server::init(bool offln)
 {
     virtual_directory = SERVER_VIRTUAL_DIR;
+    offline = offln;
 
     serverUDN = config->getOption(CFG_SERVER_UDN);
     aliveAdvertisementInterval = config->getIntOption(CFG_SERVER_ALIVE_INTERVAL);
@@ -77,21 +104,20 @@ void Server::init()
 
     // initialize what is needed
     auto self = shared_from_this();
-    timer = std::make_shared<Timer>(config);
-
-    clients = std::make_shared<Clients>(config);
-    mime = std::make_shared<Mime>(config);
+    timer = std::make_shared<Timer>();
     timer->run();
+
+    mime = std::make_shared<Mime>(config);
     database = Database::createInstance(config, mime, timer);
     config->updateConfigFromDatabase(database);
-    session_manager = std::make_shared<web::SessionManager>(config, timer);
-    context = std::make_shared<Context>(config, clients, mime, database, self, session_manager);
+
+    clientManager = std::make_shared<ClientManager>(config, database);
+    session_manager = std::make_shared<Web::SessionManager>(config, timer);
+    context = std::make_shared<Context>(config, clientManager, mime, database, session_manager);
 
     content = std::make_shared<ContentManager>(context, self, timer);
 }
 
-Server::~Server() { log_debug("Server destroyed"); }
-
 void Server::run()
 {
     log_debug("Starting...");
@@ -111,30 +137,35 @@ void Server::run()
         throw_std_runtime_error("You cannot specify interface {} and IP {} at the same time", iface, ip);
 
     if (iface.empty())
-        iface = ipToInterface(ip);
+        iface = GrbNet::ipToInterface(ip);
 
     if (!ip.empty() && iface.empty())
-        throw_std_runtime_error("Could not find IP: {}", ip.c_str());
+        throw_std_runtime_error("Could not find IP: {}", ip);
 
     // check webroot directory
-    std::string web_root = config->getOption(CFG_SERVER_WEBROOT);
-    if (web_root.empty()) {
-        throw_std_runtime_error("Invalid web server root directory {}", web_root);
+    std::string webRoot = config->getOption(CFG_SERVER_WEBROOT);
+    if (webRoot.empty()) {
+        throw_std_runtime_error("Invalid web server root directory {}", webRoot);
     }
 
-    auto ret = startupInterface(iface, in_port_t(config->getIntOption(CFG_SERVER_PORT)));
+    auto configPort = in_port_t(config->getIntOption(CFG_SERVER_PORT));
+    auto ret = startupInterface(iface, configPort);
     if (ret != UPNP_E_SUCCESS) {
-        throw UpnpException(ret, fmt::format("run: UpnpInit failed {} {}", iface, port));
+        throw UpnpException(ret, fmt::format("UpnpInit failed {} {}", iface, port));
     }
 
-    ret = UpnpSetWebServerRootDir(web_root.c_str());
+    if (configPort > 0 && port != configPort) {
+        throw UpnpException(ret, fmt::format("LibUPnP failed to bind to request port! Bound to {}, requested: {}", port, configPort));
+    }
+
+    ret = UpnpSetWebServerRootDir(webRoot.c_str());
     if (ret != UPNP_E_SUCCESS) {
-        throw UpnpException(ret, fmt::format("run: UpnpSetWebServerRootDir failed {}", web_root));
+        throw UpnpException(ret, fmt::format("UpnpSetWebServerRootDir failed {}", webRoot));
     }
 
-    log_debug("webroot: {}", web_root.c_str());
+    log_debug("webroot: {}", webRoot);
+    log_debug("Setting virtual dir to: {}", virtual_directory);
 
-    log_debug("Setting virtual dir to: {}", virtual_directory.c_str());
     ret = UpnpAddVirtualDir(virtual_directory.c_str(), this, nullptr);
     if (ret != UPNP_E_SUCCESS) {
         throw UpnpException(ret, fmt::format("run: UpnpAddVirtualDir failed {}", virtual_directory));
@@ -149,22 +180,22 @@ void Server::run()
     std::string presentationURL = getPresentationUrl();
 
     log_debug("Creating UpnpXMLBuilder");
-    xmlbuilder = std::make_unique<UpnpXMLBuilder>(context, virtualUrl, presentationURL);
+    xmlBuilder = std::make_shared<UpnpXMLBuilder>(context, getVirtualUrl(), presentationURL);
 
     // register root device with the library
-    auto desc = xmlbuilder->renderDeviceDescription();
+    auto desc = xmlBuilder->renderDeviceDescription();
     std::ostringstream buf;
     desc->print(buf, "", 0);
     std::string deviceDescription = buf.str();
-    //log_debug("Device Description: {}", deviceDescription.c_str());
+    // log_debug("Device Description: {}", deviceDescription.c_str());
 
     log_debug("Registering with UPnP...");
     ret = UpnpRegisterRootDevice2(
         UPNPREG_BUF_DESC,
         deviceDescription.c_str(),
-        size_t(deviceDescription.length()) + 1,
+        deviceDescription.length() + 1,
         true,
-        handleUpnpRootDeviceEventCallback,
+        [](auto eventType, auto event, auto cookie) { return static_cast<Server*>(cookie)->handleUpnpRootDeviceEvent(eventType, event); },
         this,
         &rootDeviceHandle);
     if (ret != UPNP_E_SUCCESS) {
@@ -172,7 +203,7 @@ void Server::run()
     }
 
     ret = UpnpRegisterClient(
-        handleUpnpClientEventCallback,
+        [](auto eventType, auto event, auto cookie) { return static_cast<Server*>(cookie)->handleUpnpClientEvent(eventType, event); },
         this,
         &clientHandle);
     if (ret != UPNP_E_SUCCESS) {
@@ -180,32 +211,48 @@ void Server::run()
     }
 
     log_debug("Creating ContentDirectoryService");
-    cds = std::make_unique<ContentDirectoryService>(context, xmlbuilder.get(), rootDeviceHandle,
+    cds = std::make_unique<ContentDirectoryService>(context, xmlBuilder, rootDeviceHandle,
         config->getIntOption(CFG_SERVER_UPNP_TITLE_AND_DESC_STRING_LIMIT));
 
     log_debug("Creating ConnectionManagerService");
-    cmgr = std::make_unique<ConnectionManagerService>(context, xmlbuilder.get(), rootDeviceHandle);
+    cmgr = std::make_unique<ConnectionManagerService>(context, xmlBuilder, rootDeviceHandle);
 
     log_debug("Creating MRRegistrarService");
-    mrreg = std::make_unique<MRRegistrarService>(context, xmlbuilder.get(), rootDeviceHandle);
+    mrreg = std::make_unique<MRRegistrarService>(context, xmlBuilder, rootDeviceHandle);
 
     // The advertisement will be sent by LibUPnP every (A/2)-30 seconds, and will have a cache-control max-age of A where A is
     // the value configured here. Ex: A value of 62 will result in an SSDP advertisement being sent every second.
-    log_debug("Sending UPnP Alive advertisements every {} seconds", (aliveAdvertisementInterval / 2) - 30);
+    log_info("Will send UPnP Alive advertisements every {} seconds", (aliveAdvertisementInterval / 2) - 30);
     ret = UpnpSendAdvertisement(rootDeviceHandle, aliveAdvertisementInterval);
     if (ret != UPNP_E_SUCCESS) {
         throw UpnpException(ret, fmt::format("run: UpnpSendAdvertisement {} failed", aliveAdvertisementInterval));
     }
 
-    std::string url = config->getOption(CFG_VIRTUAL_URL);
-    if (url.empty()) {
-        url = renderWebUri(ip, port);
-    }
-    if (url.find("http") != 0) { // url does not start with http
-        url = fmt::format("http://{}", url);
-    }
+    UpnpSetHostValidateCallback(
+        [](auto host, auto cookie) -> int {
+            auto hostStr = std::string(host);
+            auto ip = std::string(UpnpGetServerIpAddress());
+            auto ip6 = std::string(UpnpGetServerIp6Address());
+            auto ip6gla = std::string(UpnpGetServerUlaGuaIp6Address());
+
+            if (hostStr.find(ip) != std::string::npos || hostStr.find(ip6) != std::string::npos || hostStr.find(ip6gla) != std::string::npos) {
+                return UPNP_E_SUCCESS;
+            }
+
+            auto virtualURL = static_cast<Server*>(cookie)->getVirtualUrl();
+            if (!virtualURL.empty() && virtualURL.find(host) != std::string::npos) {
+                return UPNP_E_SUCCESS;
+            }
+            log_warning("Rejected attempt to load host '{}' as it does not match configured virtualURL: '{}'. "
+                        "See https://docs.gerbera.io/en/stable/config-server.html#virtualurl",
+                host, static_cast<Server*>(cookie)->config->getOption(CFG_VIRTUAL_URL));
+            return UPNP_E_BAD_HTTPMSG;
+        },
+        this);
+
+    std::string url = getVirtualUrl();
     writeBookmark(url);
-    log_info("The Web UI can be reached by following this link: {}/", url);
+    log_info("The Web UI can be reached by following this link: {}", url);
 }
 
 int Server::startupInterface(const std::string& iface, in_port_t inPort)
@@ -215,19 +262,21 @@ int Server::startupInterface(const std::string& iface, in_port_t inPort)
     const char* ifName = iface.empty() ? nullptr : iface.c_str();
 
     int ret = UPNP_E_INIT_FAILED;
-    for (size_t attempt = 0; ret != UPNP_E_SUCCESS; attempt++) {
+    for (std::size_t attempt = 0; attempt < 4; attempt++) {
         ret = UpnpInit2(ifName, inPort);
-        if (ret != UPNP_E_SUCCESS) {
-            if (attempt > 3) {
-                throw UpnpException(ret, fmt::format("run: UpnpInit failed with {} {}", ifName, inPort));
-            }
+        if (ret == UPNP_E_SUCCESS)
+            break;
+
 #ifdef UPNP_HAVE_TOOLS
-            log_warning("UPnP Init {}:{} failed: {} ({}). Retrying in {} seconds...", ifName, inPort, UpnpGetErrorMessage(ret), ret, attempt + 1);
+        log_warning("UPnP Init {}:{} failed: {} ({}). Retrying in {} seconds...", ifName, inPort, UpnpGetErrorMessage(ret), ret, attempt + 1);
 #else
-            log_warning("UPnP Init {}:{} failed: ({}). Retrying in {} seconds...", ifName, inPort, ret, attempt + 1);
+        log_warning("UPnP Init {}:{} failed: ({}). Retrying in {} seconds...", ifName, inPort, ret, attempt + 1);
 #endif
-            std::this_thread::sleep_for(std::chrono::seconds(attempt + 1));
-        }
+        std::this_thread::sleep_for(std::chrono::seconds(attempt + 1));
+    }
+
+    if (ret != UPNP_E_SUCCESS) {
+        throw UpnpException(ret, fmt::format("run: UpnpInit failed with {} {}", ifName, inPort));
     }
 
     port = UpnpGetServerPort();
@@ -238,8 +287,6 @@ int Server::startupInterface(const std::string& iface, in_port_t inPort)
     log_info("IPv6: Server bound to: {}:{}", UpnpGetServerIp6Address(), UpnpGetServerPort6());
     log_info("IPv6 ULA/GLA: Server bound to: {}:{}", UpnpGetServerUlaGuaIp6Address(), UpnpGetServerUlaGuaPort6());
 
-    virtualUrl = fmt::format("http://{}:{}/{}", ip, port, virtual_directory);
-
     return ret;
 }
 
@@ -267,22 +314,30 @@ void Server::writeBookmark(const std::string& addr)
 
     fs::path path = config->getOption(CFG_SERVER_BOOKMARK_FILE);
     log_debug("Writing bookmark file to: {}", path.c_str());
-    writeTextFile(path, data);
+    GrbFile(std::move(path)).writeTextFile(data);
 }
 
 void Server::emptyBookmark()
 {
-    const std::string data = "<html><body><h1>Gerbera Media Server is not running.</h1><p>Please start it and try again.</p></body></html>";
+    const std::string_view data = "<html><body><h1>Gerbera Media Server is not running.</h1><p>Please start it and try again.</p></body></html>";
 
     fs::path path = config->getOption(CFG_SERVER_BOOKMARK_FILE);
     log_debug("Clearing bookmark file at: {}", path.c_str());
-    writeTextFile(path, data);
+    GrbFile(std::move(path)).writeTextFile(data);
 }
 
 std::string Server::getVirtualUrl() const
 {
-    auto cfgVirt = config->getOption(CFG_VIRTUAL_URL);
-    return cfgVirt.empty() ? virtualUrl : fmt::format("{}/{}", cfgVirt, virtual_directory);
+    auto virtUrl = config->getOption(CFG_VIRTUAL_URL);
+    if (virtUrl.empty()) {
+        virtUrl = GrbNet::renderWebUri(ip, port);
+    }
+    if (!startswith(virtUrl, "http")) { // url does not start with http
+        virtUrl = fmt::format("http://{}", virtUrl);
+    }
+    if (virtUrl.back() == '/')
+        virtUrl.pop_back();
+    return virtUrl;
 }
 
 bool Server::getShutdownStatus() const
@@ -335,13 +390,10 @@ void Server::shutdown()
     timer->shutdown();
     timer = nullptr;
 
-    mime = nullptr;
-    clients = nullptr;
-}
+    context = nullptr;
 
-int Server::handleUpnpRootDeviceEventCallback(Upnp_EventType eventType, const void* event, void* cookie)
-{
-    return static_cast<Server*>(cookie)->handleUpnpRootDeviceEvent(eventType, event);
+    mime = nullptr;
+    clientManager = nullptr;
 }
 
 int Server::handleUpnpRootDeviceEvent(Upnp_EventType eventType, const void* event)
@@ -361,11 +413,11 @@ int Server::handleUpnpRootDeviceEvent(Upnp_EventType eventType, const void* even
     case UPNP_CONTROL_ACTION_REQUEST:
         log_debug("UPNP_CONTROL_ACTION_REQUEST");
         try {
-            auto request = std::make_unique<ActionRequest>(context, static_cast<UpnpActionRequest*>(const_cast<void*>(event)));
+            auto request = ActionRequest(xmlBuilder, clientManager, static_cast<UpnpActionRequest*>(const_cast<void*>(event)));
             routeActionRequest(request);
-            request->update();
-        } catch (const UpnpException& upnp_e) {
-            ret = upnp_e.getErrorCode();
+            request.update();
+        } catch (const UpnpException& upnpE) {
+            ret = upnpE.getErrorCode();
             UpnpActionRequest_set_ErrCode(static_cast<UpnpActionRequest*>(const_cast<void*>(event)), ret);
         } catch (const std::runtime_error& e) {
             log_info("Exception: {}", e.what());
@@ -375,11 +427,11 @@ int Server::handleUpnpRootDeviceEvent(Upnp_EventType eventType, const void* even
     case UPNP_EVENT_SUBSCRIPTION_REQUEST:
         log_debug("UPNP_EVENT_SUBSCRIPTION_REQUEST");
         try {
-            auto request = std::make_unique<SubscriptionRequest>(static_cast<UpnpSubscriptionRequest*>(const_cast<void*>(event)));
+            auto request = SubscriptionRequest(static_cast<UpnpSubscriptionRequest*>(const_cast<void*>(event)));
             routeSubscriptionRequest(request);
-        } catch (const UpnpException& upnp_e) {
-            log_warning("Subscription exception: {}", upnp_e.what());
-            ret = upnp_e.getErrorCode();
+        } catch (const UpnpException& upnpE) {
+            log_warning("Subscription exception: {}", upnpE.what());
+            ret = upnpE.getErrorCode();
         }
         break;
 
@@ -394,11 +446,6 @@ int Server::handleUpnpRootDeviceEvent(Upnp_EventType eventType, const void* even
     return ret;
 }
 
-int Server::handleUpnpClientEventCallback(Upnp_EventType eventType, const void* event, void* cookie)
-{
-    return static_cast<Server*>(cookie)->handleUpnpClientEvent(eventType, event);
-}
-
 int Server::handleUpnpClientEvent(Upnp_EventType eventType, const void* event)
 {
     // check parameters
@@ -410,12 +457,12 @@ int Server::handleUpnpClientEvent(Upnp_EventType eventType, const void* event)
     switch (eventType) {
     case UPNP_DISCOVERY_ADVERTISEMENT_ALIVE:
     case UPNP_DISCOVERY_SEARCH_RESULT: {
-        auto d_event = reinterpret_cast<const UpnpDiscovery*>(event);
-        const char* userAgent = UpnpDiscovery_get_Os_cstr(d_event);
-        const struct sockaddr_storage* destAddr = UpnpDiscovery_get_DestAddr(d_event);
-        const char* location = UpnpDiscovery_get_Location_cstr(d_event);
+        auto dEvent = static_cast<const UpnpDiscovery*>(event);
+        const char* userAgent = UpnpDiscovery_get_Os_cstr(dEvent);
+        auto destAddr = std::make_shared<GrbNet>(UpnpDiscovery_get_DestAddr(dEvent));
+        const char* location = UpnpDiscovery_get_Location_cstr(dEvent);
 
-        clients->addClientByDiscovery(destAddr, userAgent, location);
+        clientManager->addClientByDiscovery(destAddr, userAgent, location);
         break;
     }
     default:
@@ -425,26 +472,30 @@ int Server::handleUpnpClientEvent(Upnp_EventType eventType, const void* event)
     return 0;
 }
 
-void Server::routeActionRequest(const std::unique_ptr<ActionRequest>& request) const
+void Server::routeActionRequest(ActionRequest& request)
 {
     log_debug("start");
 
     // make sure the request is for our device
-    if (request->getUDN() != serverUDN) {
+    if (request.getUDN() != serverUDN) {
         // not for us
         throw UpnpException(UPNP_E_BAD_REQUEST, "routeActionRequest: request not for this device");
     }
 
     // we need to match the serviceID to one of our services
-    if (request->getServiceID() == UPNP_DESC_CM_SERVICE_ID) {
+    if (request.getServiceID() == UPNP_DESC_CM_SERVICE_ID) {
         // this call is for the lifetime stats service
         // log_debug("request for connection manager service");
         cmgr->processActionRequest(request);
-    } else if (request->getServiceID() == UPNP_DESC_CDS_SERVICE_ID) {
+    } else if (request.getServiceID() == UPNP_DESC_CDS_SERVICE_ID) {
         // this call is for the toaster control service
-        //log_debug("routeActionRequest: request for content directory service");
-        cds->processActionRequest(request);
-    } else if (request->getServiceID() == UPNP_DESC_MRREG_SERVICE_ID) {
+        // log_debug("routeActionRequest: request for content directory service");
+        if (!offline)
+            cds->processActionRequest(request);
+        else
+            request.setErrorCode(UPNP_E_INVALID_ACTION);
+
+    } else if (request.getServiceID() == UPNP_DESC_MRREG_SERVICE_ID) {
         mrreg->processActionRequest(request);
     } else {
         // cp is asking for a nonexistent service, or for a service
@@ -453,26 +504,25 @@ void Server::routeActionRequest(const std::unique_ptr<ActionRequest>& request) c
     }
 }
 
-void Server::routeSubscriptionRequest(const std::unique_ptr<SubscriptionRequest>& request) const
+void Server::routeSubscriptionRequest(const SubscriptionRequest& request) const
 {
     // make sure that the request is for our device
-    if (request->getUDN() != serverUDN) {
+    if (request.getUDN() != serverUDN) {
         // not for us
-        log_debug("routeSubscriptionRequest: request not for this device: {} vs {}",
-            request->getUDN().c_str(), serverUDN.c_str());
-        throw UpnpException(UPNP_E_BAD_REQUEST, "routeActionRequest: request not for this device");
+        log_debug("routeSubscriptionRequest: request not for this device: {} vs {}", request.getUDN(), serverUDN);
+        throw UpnpException(UPNP_E_BAD_REQUEST, "routeSubscriptionRequest: request not for this device");
     }
 
     // we need to match the serviceID to one of our services
-    if (request->getServiceID() == UPNP_DESC_CDS_SERVICE_ID) {
+    if (request.getServiceID() == UPNP_DESC_CDS_SERVICE_ID) {
         // this call is for the content directory service
-        //log_debug("routeSubscriptionRequest: request for content directory service");
+        // log_debug("routeSubscriptionRequest: request for content directory service");
         cds->processSubscriptionRequest(request);
-    } else if (request->getServiceID() == UPNP_DESC_CM_SERVICE_ID) {
+    } else if (request.getServiceID() == UPNP_DESC_CM_SERVICE_ID) {
         // this call is for the connection manager service
-        //log_debug("routeSubscriptionRequest: request for connection manager service");
+        // log_debug("routeSubscriptionRequest: request for connection manager service");
         cmgr->processSubscriptionRequest(request);
-    } else if (request->getServiceID() == UPNP_DESC_MRREG_SERVICE_ID) {
+    } else if (request.getServiceID() == UPNP_DESC_MRREG_SERVICE_ID) {
         mrreg->processSubscriptionRequest(request);
     } else {
         // cp asks for a nonexistent service or for a service that
@@ -489,37 +539,26 @@ void Server::sendCDSSubscriptionUpdate(const std::string& updateString)
 
 std::unique_ptr<RequestHandler> Server::createRequestHandler(const char* filename) const
 {
-    std::string link = urlUnescape(filename);
+    std::string link = URLUtils::urlUnescape(filename);
     log_debug("Filename: {}", filename);
 
     if (startswith(link, fmt::format("/{}/{}", SERVER_VIRTUAL_DIR, CONTENT_MEDIA_HANDLER))) {
-        return std::make_unique<FileRequestHandler>(content, xmlbuilder.get());
+        return std::make_unique<FileRequestHandler>(content, xmlBuilder);
     }
 
     if (startswith(link, fmt::format("/{}/{}", SERVER_VIRTUAL_DIR, CONTENT_UI_HANDLER))) {
-        std::string parameters;
-        std::string path;
-        RequestHandler::splitUrl(filename, URL_UI_PARAM_SEPARATOR, path, parameters);
+        auto&& [path, parameters] = URLUtils::splitUrl(filename, URL_UI_PARAM_SEPARATOR);
 
-        std::map<std::string, std::string> params;
-        dictDecode(parameters, &params);
+        auto params = URLUtils::dictDecode(parameters);
 
         auto it = params.find(URL_REQUEST_TYPE);
-        std::string r_type = it != params.end() && !it->second.empty() ? it->second : "index";
+        std::string rType = it != params.end() && !it->second.empty() ? it->second : "index";
 
-        return web::createWebRequestHandler(content, r_type);
+        return Web::createWebRequestHandler(context, content, xmlBuilder, rType);
     }
 
     if (startswith(link, fmt::format("/{}/{}", SERVER_VIRTUAL_DIR, DEVICE_DESCRIPTION_PATH))) {
-        return std::make_unique<DeviceDescriptionHandler>(content, xmlbuilder.get());
-    }
-
-    if (startswith(link, fmt::format("/{}/{}", SERVER_VIRTUAL_DIR, CONTENT_SERVE_HANDLER))) {
-        if (config->getOption(CFG_SERVER_SERVEDIR).empty()) {
-            throw_std_runtime_error("Serving directories is not enabled in configuration");
-        }
-
-        return std::make_unique<ServeRequestHandler>(content);
+        return std::make_unique<DeviceDescriptionHandler>(content, xmlBuilder);
     }
 
 #if defined(HAVE_CURL)
@@ -537,10 +576,10 @@ int Server::registerVirtualDirCallbacks()
     int ret = UpnpVirtualDir_set_GetInfoCallback([](const char* filename, UpnpFileInfo* info, const void* cookie, const void** requestCookie) -> int {
         try {
             auto reqHandler = static_cast<const Server*>(cookie)->createRequestHandler(filename);
-            std::string link = urlUnescape(filename);
+            std::string link = URLUtils::urlUnescape(filename);
             reqHandler->getInfo(startswith(link, fmt::format("/{}/{}", SERVER_VIRTUAL_DIR, CONTENT_UI_HANDLER)) ? filename : link.c_str(), info);
             return 0;
-        } catch (const ServerShutdownException& se) {
+        } catch (const ServerShutdownException&) {
             return -1;
         } catch (const SubtitlesNotFoundException& sex) {
             log_warning("{}", sex.what());
@@ -557,11 +596,14 @@ int Server::registerVirtualDirCallbacks()
     ret = UpnpVirtualDir_set_OpenCallback([](const char* filename, enum UpnpOpenFileMode mode, const void* cookie, const void* requestCookie) -> UpnpWebFileHandle {
         try {
             auto reqHandler = static_cast<const Server*>(cookie)->createRequestHandler(filename);
-            std::string link = urlUnescape(filename);
+            std::string link = URLUtils::urlUnescape(filename);
             auto ioHandler = reqHandler->open(startswith(link, fmt::format("/{}/{}", SERVER_VIRTUAL_DIR, CONTENT_UI_HANDLER)) ? filename : link.c_str(), mode);
-            auto ioPtr = UpnpWebFileHandle(ioHandler.release());
-            //log_debug("{} open({})", ioPtr, filename);
-            return ioPtr;
+            if (ioHandler) {
+                ioHandler->open(mode);
+                return ioHandler.release();
+            }
+            log_warning("No Handler for {}", link);
+            return nullptr;
         } catch (const ServerShutdownException& se) {
             return nullptr;
         } catch (const SubtitlesNotFoundException& sex) {
@@ -576,20 +618,20 @@ int Server::registerVirtualDirCallbacks()
         return ret;
 
     log_debug("Setting UpnpVirtualDir ReadCallback");
-    ret = UpnpVirtualDir_set_ReadCallback([](UpnpWebFileHandle f, char* buf, size_t length, const void* cookie, const void* requestCookie) -> int {
-        //log_debug("{} read({})", f, length);
+    ret = UpnpVirtualDir_set_ReadCallback([](UpnpWebFileHandle f, char* buf, std::size_t length, const void* cookie, const void* requestCookie) -> int {
+        // log_debug("{} read({})", f, length);
         if (static_cast<const Server*>(cookie)->getShutdownStatus())
             return -1;
 
-        auto handler = static_cast<IOHandler*>(f);
-        return handler->read(buf, length);
+        auto ioHandler = static_cast<IOHandler*>(f);
+        return ioHandler ? ioHandler->read(reinterpret_cast<std::byte*>(buf), length) : 0;
     });
     if (ret != UPNP_E_SUCCESS)
         return ret;
 
     log_debug("Setting UpnpVirtualDir WriteCallback");
-    ret = UpnpVirtualDir_set_WriteCallback([](UpnpWebFileHandle f, char* buf, size_t length, const void* cookie, const void* requestCookie) -> int {
-        //log_debug("{} write({})", f, length);
+    ret = UpnpVirtualDir_set_WriteCallback([](UpnpWebFileHandle f, char* buf, std::size_t length, const void* cookie, const void* requestCookie) -> int {
+        // log_debug("{} write({})", f, length);
         return 0;
     });
     if (ret != UPNP_E_SUCCESS)
@@ -597,10 +639,11 @@ int Server::registerVirtualDirCallbacks()
 
     log_debug("Setting UpnpVirtualDir SeekCallback");
     ret = UpnpVirtualDir_set_SeekCallback([](UpnpWebFileHandle f, off_t offset, int whence, const void* cookie, const void* requestCookie) -> int {
-        //log_debug("{} seek({}, {})", f, offset, whence);
+        // log_debug("{} seek({}, {})", f, offset, whence);
         try {
-            auto handler = static_cast<IOHandler*>(f);
-            handler->seek(offset, whence);
+            auto ioHandler = static_cast<IOHandler*>(f);
+            if (ioHandler)
+                ioHandler->seek(offset, whence);
             return 0;
         } catch (const std::runtime_error& e) {
             log_error("Exception during seek: {}", e.what());
@@ -612,17 +655,19 @@ int Server::registerVirtualDirCallbacks()
 
     log_debug("Setting UpnpVirtualDir CloseCallback");
     ret = UpnpVirtualDir_set_CloseCallback([](UpnpWebFileHandle f, const void* cookie, const void* requestCookie) -> int {
-        //log_debug("{} close()", f);
-        int ret_close = 0;
-        auto handler = std::unique_ptr<IOHandler>(static_cast<IOHandler*>(f));
-        try {
-            handler->close();
-        } catch (const std::runtime_error& e) {
-            log_error("Exception during close: {}", e.what());
-            ret_close = -1;
+        // log_debug("{} close()", f);
+        int retClose = 0;
+        auto ioHandler = std::unique_ptr<IOHandler>(static_cast<IOHandler*>(f));
+        if (ioHandler) {
+            try {
+                ioHandler->close();
+            } catch (const std::runtime_error& e) {
+                log_error("Exception during close: {}", e.what());
+                retClose = -1;
+            }
         }
 
-        return ret_close;
+        return retClose;
     });
 
     return ret;
